#ifndef JTHREAD_H__
#define JTHREAD_H__

#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
#endif
#include <jni.h>
#include <jvmpi.h>
#include <timerstack.h>
#include <hash.h>
#include <jmp.h>
#include <cls.h>
#include <method.h>

#define THREAD_T_UNKNOWN	0x00
#define THREAD_T_SYSTEM		0x01
#define THREAD_T_JMP		0x02
#define THREAD_T_APPLICATION	0x04
#define THREAD_F_INTERESTED	0x80

/** This is the structure used for a java thread. */ 
struct jmpthread {
    char* thread_name;       /** The name of the thread. */
    char* group_name;        /** The name of the thread group. */
    char* parent_name;       /** The name of the parent thread. */
    jobjectID thread_id;     /** The id of the thread. */
    JNIEnv* env_id;          /** The environment id of the thread. */
    jint state;              /** The last known state of the thread. */
    timerstack* timerstack;
    unsigned char mode;      /** Thread clasification and basic filter */
};

/** Create a new thread structure with the given attributes. */
jmpthread* jmpthread_new (const char* thread_name, const char* group_name, 
		      const char* parent_name, jobjectID thread_id,
		      JNIEnv* env_id, timerstack *s, unsigned char mode);

/** Delete the given thread structure. */
void jmpthread_free (jmpthread* t);

/** The hash function for threads. */
size_t jmpthread_jmphash_func (void* c, size_t len);

/** Check if two threads are equal.
 *  return 0 if the threads are equal.
 */
int jmpthread_cmp_func (void* c1, void* c2);

/** Get the environment id of the given thread. */
JNIEnv* jmpthread_get_env_id (jmpthread* t);

/** Set the environment id of the given thread. */
void jmpthread_set_env_id (jmpthread* t, JNIEnv* env_id);

/** The given thread has entered a method at time tv. 
 */
void jmpthread_method_entry (timerstack* s, method *m, jlong tval);

/** The given thread has exited from a method.
 */
void jmpthread_method_exit(timerstack* s, jmethodID method_id,
			 jlong tval, JNIEnv* env);

/** The given thread has is waiting for a monitor. */
void jmpthread_contenation_enter (timerstack* s,
				obj* monitor,
				jlong tval);

/** The given thread got the monitor it was waiting for. */
void jmpthread_contenation_entered (timerstack* s,
				  obj* monitor, 
				  jlong tval);

/** Get the name of the given thread. */
char* jmpthread_get_thread_name (jmpthread* t);

/** Get the name of the group for the given thread. */
char* jmpthread_get_group_name (jmpthread* t);

/** Get the name of the parent for the given thread. */
char* jmpthread_get_parent_name (jmpthread* t);

/** Get thread filter mode info from env_id */
unsigned char jmpthread_get_mode_by_env_id (JNIEnv* env_id);

#endif   /* JTHREAD_H__ */

/* Emacs Local Variables: */
/* Emacs mode:C */
/* Emacs c-indentation-style:"gnu" */
/* Emacs c-hanging-braces-alist:((brace-list-open)(brace-entry-open)(defun-open after)(substatement-open after)(block-close . c-snug-do-while)(extern-lang-open after)) */
/* Emacs c-cleanup-list:(brace-else-brace brace-elseif-brace space-before-funcall) */
/* Emacs c-basic-offset:4 */
/* Emacs End: */
